#
#   Pyrex Top Level
#

import os
import sys
from Scanning import PyrexScanner
import Errors
from Errors import PyrexError, CompileError, InternalError
import Parsing
import Code
from Pyrex.Utils import replace_suffix

if sys.platform == "mac":
  from Pyrex.Mac.MacSystem import c_compile, c_link

class CompilationResult:
  #  c_file           string or None
  #  h_file           string or None
  #  listing_file     string or None
  #  object_file      string or None
  #  extension_file   string or None
  #  num_errors       integer
  #
  #  Used internally:
  #
  #  c_file_generated   boolean
  #  h_file_generated   boolean
  
  def __init__(self):
    self.c_file = None
    self.h_file = None
    self.listing_file = None
    self.object_file = None
    self.extension_file = None
    self.c_file_generated = 0
    self.h_file_generated = 0


def extract_module_name(path):
  _, tail = os.path.split(path)
  name, _ = os.path.splitext(tail)
  return name

def parse(source):
  f = open(source, "r")
  s = PyrexScanner(f, source)
  try:
    tree = Parsing.p_module(s)
  finally:
    f.close()
  return tree

#def generate(name, tree, output):
#	code = Code.Generator(output)
#	tree.process(name, code)

def compile(source, 
    keep_files = 0,
    c_only = 0,
    obj_only = 0,
    dump_parse_tree = 0,
    use_listing_file = 1):
  try:
    result = CompilationResult()
    output_generated = 0
    source = os.path.join(os.getcwd(), source)
    if use_listing_file:
      result.listing_file = replace_suffix(source, ".lis")
      Errors.open_listing_file(result.listing_file)
    try:
      tree = parse(source)
      if dump_parse_tree:
        Parsing.print_parse_tree(sys.stdout, tree, 0)
      result.c_file = replace_suffix(source, ".c")
      result.h_file = replace_suffix(source, ".h")
      name = extract_module_name(source)
      #generate(name, tree, result)
      tree.process(name, result)
      output_generated = 1
    finally:
      Errors.close_listing_file()
  except CompileError:
    pass
  if not result.h_file_generated:
    result.h_file = None
  if not result.c_file_generated:
    result.c_file = None
  result.num_errors = Errors.num_errors
  if result.num_errors > 0 or not output_generated:
    return result
  if c_only:
    return result
  result.object_file = c_compile(result.c_file)
  if obj_only:
    return result
  result.extension_file = c_link(result.object_file)
  return result

def main(c_only = 0, use_listing_file = 1):
  any_failures = 0
  for arg in sys.argv[1:]:
    try:
      result = compile(arg, 
        c_only = c_only, 
        use_listing_file = use_listing_file)
      if result.num_errors > 0:
        any_failures = 1
    except PyrexError, e:
      print e
      any_failures = 1
  if any_failures:
    sys.exit(1)


